/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2006 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include <stdlib.h>

#include "sambascanner.h"
#include "String.h"
#include "Parser.h"
#include "CommandLine.h"

void Display_Help(char **argv) {
	fprintf(stderr, "Syntax: %s [Options]\n", argv[0]);
	fprintf(stderr, "\n");
	fprintf(stderr, "-t, --time       Requires a float argument which states by which factor the\n");
	fprintf(stderr, "                 internal timeouts should be multiplied with.\n");
	fprintf(stderr, "-w, --workgroup  Manually sets the workgroup to the given argument.\n");
	fprintf(stderr, "-d, --debug      Set the debug level from 0 (silent) to 3 (full debug),\n");
	fprintf(stderr, "                 default level is 1.\n");
	fprintf(stderr, "-s, --smbclient  Set the path where smbclient can be found (default " DEFAULT_SMBCLIENT ").\n");
	fprintf(stderr, "-n, --nmblookup  Set the path where nmblookup can be found (default " DEFAULT_NMBLOOKUP ").\n");
	fprintf(stderr, "-l, --smblister  Set the path where smblister can be found (default " DEFAULT_SMBLISTER ").\n");
	fprintf(stderr, "-c, --computer   Specify a computer on the network.\n");
	fprintf(stderr, "-o, --outfile    Set the output file (default 'Filelist.txt').\n");
	fprintf(stderr, "-r, --threadcnt  Specifies how many threads should be used (default 30).\n");
	fprintf(stderr, "-p, --preserve   Doesn't delete tempfiles after work is done.\n");
	fprintf(stderr, "-h, --help       Display this help.\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "Example: %s -w MYGROUP -t 2.5\n", argv[0]);
	fprintf(stderr, "         Multiplies all timings by 2.5 and sets the workgroup to 'MYGROUP'.\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "         %s -w MATRIX -c NEO\n", argv[0]);
	fprintf(stderr, "         Select workgroup 'MYGROUP' and specify known computer 'NEO'.\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "This is SambaScanner " VERSION ", released under the GNU GPL-2.\n");
}

int Process_Parameters(int argc, char **argv, StringList *Hostlist) {
	char *Equivalent_Option = NULL;

	/* Default options */
	Options.Workgroup[0] = 0;
	Options.Timings = 1.0;
	Options.DebugLvl = 1;
	Options.ThreadCount = 30;
	Options.PreserveTemp = 0;
	strcpy(Options.SmbClient_Executable, DEFAULT_SMBCLIENT);
	strcpy(Options.NmbLookup_Executable, DEFAULT_NMBLOOKUP);
	strcpy(Options.SmbLister_Executable, DEFAULT_SMBLISTER);
	strcpy(Options.Outfile, "Filelist.txt");

	while (1) {
		int Result;
		int Option_Index = 0;
		static struct option Long_Options[] = {
			{"time", 1, 0, 0},
			{"workgroup", 1, 0, 0},
			{"help", 0, 0, 0},
			{"debug", 1, 0, 0},
			{"smbclient", 1, 0, 0},
			{"nmblookup", 1, 0, 0},
			{"smblister", 1, 0, 0},
			{"outfile", 1, 0, 0},
			{"computer", 1, 0, 0},
			{"threadcnt", 1, 0, 0},
			{"preserve", 0, 0, 0},
			{0, 0, 0, 0}
		};
		const char *Short_Options = "t:w:hd:s:n:l:o:c:r:p";
		Result = getopt_long (argc, argv, Short_Options, Long_Options, &Option_Index);
		if (Result == -1) break;

		if (Result == 0) {
			int i, j;
			if (Equivalent_Option == NULL) {
				Equivalent_Option = (char*)malloc(strlen(Short_Options)+1);
				if (!Equivalent_Option) {
					perror("malloc Equivalent_Option");
					exit(EXIT_FAILURE);
				}
				
				j = 0;
				for (i = 0; i < strlen(Short_Options); i++) {
					if (Short_Options[i] != ':') {
						Equivalent_Option[j] = Short_Options[i];
						j++;
					}
				}
				Equivalent_Option[j] = 0;
			}
			Result = Equivalent_Option[Option_Index];
		}
		
		switch (Result) {
			case 't': {
				double Timeout;
				Timeout = atof(optarg);
				if ((Timeout<TIMEOUT_MIN) || (Timeout>TIMEOUT_MAX)) {
					fprintf(stderr, "Error: Timeout must be inbetween %.1f and %.1f.\n", TIMEOUT_MIN, TIMEOUT_MAX);
					return 0;
				}
				Options.Timings=Timeout;
				break;
			}
					  
			case 'w': {
				strncpy(Options.Workgroup, optarg, sizeof(Options.Workgroup));
				Options.Workgroup[sizeof(Options.Workgroup)-1] = '\0';
				break;
			}

			case 'd': {
				int DebugLvl;
				DebugLvl=atoi(optarg);
				if ((DebugLvl<0) || (DebugLvl>3)) {
					fprintf(stderr, "Error: Debuglevel must be inbetween 0 and 3.\n");
					return 0;
				}
				Options.DebugLvl=DebugLvl;
				break;
			}
			
			case 's': { /* --smbclient */
				strncpy(Options.SmbClient_Executable, optarg, sizeof(Options.SmbClient_Executable));
				Options.SmbClient_Executable[sizeof(Options.SmbClient_Executable)-1] = '\0';
				break;
			}
			case 'n': { /* --nmblookup */
				strncpy(Options.NmbLookup_Executable, optarg, sizeof(Options.NmbLookup_Executable));
				Options.NmbLookup_Executable[sizeof(Options.NmbLookup_Executable)-1] = '\0';
				break;
			}
			case 'l': { /* --smblister */
				strncpy(Options.SmbLister_Executable, optarg, sizeof(Options.SmbLister_Executable));
				Options.SmbLister_Executable[sizeof(Options.SmbLister_Executable)-1] = '\0';
				break;
			}
			case 'o': {	/* --outfile */
				strncpy(Options.Outfile, optarg, sizeof(Options.Outfile));
				Options.Outfile[sizeof(Options.Outfile)-1] = '\0';
				break;
			}
			case 'c': { /* --computer */
				Add_String(Hostlist, optarg);
				break;
			}
			case 'r': { /* --threadcnt */
				int ThreadCount;
				ThreadCount = atoi(optarg);
				if ((ThreadCount < 1) || (ThreadCount > 100)) {
					fprintf(stderr, "Error: Threadcount must stay inbetween 1 and 100\n");
					return 0;
				}
				Options.ThreadCount = ThreadCount;
				break;
			}

			case 'p': { /* --preserve */
				Options.PreserveTemp = 1;
				break;
			}
					  
			case '?':
			case 'h': {
				Display_Help(argv);
				return 0;
				break;
			}

			default: {
				fprintf(stderr, "Returned character code was 0x%x\n", Result);
				Display_Help(argv);
				return 0;
				break;
			}
		}
	}

	if (optind<argc) {
		fprintf(stderr, "Invalid argument: ");
		while (optind<argc) fprintf(stderr, "%s ", argv[optind++]);
		fprintf(stderr, "\n");
		Display_Help(argv);
		return 0;
	}

	if (Equivalent_Option) {
		free(Equivalent_Option);
	}
	return 1;
}

char *SingularPlural(char *Singular, char *Plural, int Number) {
	return (Number==1) ? Singular : Plural;
}
