/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2006 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "sambascanner.h"
#include "Mutex.h"
#include "Tools.h"
#include "String.h"
#include "Threads.h"
#include "Processes.h"

int Exec_Timeout(const char *Command, const char *Parameter, double TermTime, double KillTime, char *Outfile, int UniqueThreadID, struct Process **Processes, int *NumberProcesses) {
	pid_t PID;
	char **Arguments = NULL;
	int ArgumentCount = 0;
	char *Token;
	char *String;
	double Time;
	int CurrentProcess;
	sigset_t Mask;
					
	int Modus[2];
	struct Redirect_Thread_InputData *Data;

	message(2, "About to execute command: %s (%s) -> %s\n", Command, Parameter, Outfile);

	Data = malloc(sizeof(struct Redirect_Thread_InputData));
	if (pipe(Modus) == -1) {
		message(1, "Error creating redirection pipe for execution of '%s' (%s): %s\n", Command, Parameter, strerror(errno));
		exit(EXIT_FAILURE);
	}
						
	Lock_Exec();
	Time = GetTime();
	*Processes = (struct Process*)realloc(*Processes, sizeof(struct Process)*((*NumberProcesses) + 1));
	((*Processes)[*NumberProcesses]).Running = 1;
	((*Processes)[*NumberProcesses]).Status = 0;
	((*Processes)[*NumberProcesses]).Signal = 0;
	((*Processes)[*NumberProcesses]).Valid = 0;
	((*Processes)[*NumberProcesses]).TermTime = Time + (TermTime * Options.Timings);
	((*Processes)[*NumberProcesses]).KillTime = Time + ((TermTime + KillTime) * Options.Timings);
	(*NumberProcesses)++;
	CurrentProcess = (*NumberProcesses) - 1;
	
	Data->Outfile = Outfile;
	Data->Read_FD = Modus[0];
	Data->ProcessID = CurrentProcess;
	Data->TermTime = TermTime;
	Data->KillTime = KillTime;
	Data->Active = 0;
	Data->Processes = Processes;
	if (pthread_create(&(((*Processes)[CurrentProcess]).ThreadID), NULL, Redirect_Thread, Data)) {
		message(1, "Error creating redirection thread for execution of '%s' (%s): %s\n", Command, Parameter, strerror(errno));
		exit(1);
	}
	Unlock_Exec();	

	sigemptyset(&Mask);
	sigaddset(&Mask, SIGCHLD);
	sigprocmask(SIG_BLOCK, &Mask, NULL);
	
	PID = fork();
	if (PID == 0) {
		/* Close reading end of pipe */
		close(Modus[0]);

		/* Child process, parse arguments */
		String = (char*)malloc(strlen(Parameter)+1);
		if (String == NULL) {
			message(1, "Malloc failed when allocating string: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}

		ArgumentCount++;
		Arguments = (char**)realloc(Arguments, sizeof(char**)*ArgumentCount);
		Arguments[ArgumentCount-1] = (char*)malloc((strlen(Command)*sizeof(char))+1);
		strcpy(Arguments[ArgumentCount-1], Command);
		
		strcpy(String, Parameter);
		Token = strtok(String, " ");
		while (Token != NULL) {
			ArgumentCount++;
			Arguments = (char**)realloc(Arguments, sizeof(char**)*ArgumentCount);
			Arguments[ArgumentCount-1] = (char*)malloc((strlen(Token)*sizeof(char))+1);
			strcpy(Arguments[ArgumentCount-1], Token);
			Token = strtok(NULL, " ");
		}
		ArgumentCount++;
		Arguments = (char**)realloc(Arguments, sizeof(char**)*ArgumentCount);
		Arguments[ArgumentCount-1] = NULL;

		{
			int Old_Stdout, Old_Stderr;
			int DevNull;

			DevNull = open("/dev/null", O_WRONLY, 0644);
			if (DevNull == -1) {
				message(1, "Fork failed when opening channel to /dev/null: %s\n", strerror(errno));
				exit(EXIT_FAILURE);
			}
			
			Old_Stdout = dup(STDOUT_FILENO);
			Old_Stderr = dup(STDERR_FILENO);
			dup2(Modus[1], STDOUT_FILENO);
			dup2(DevNull, STDERR_FILENO);
			Data->Active = 1;
			execv(Command, Arguments);
			dup2(Old_Stdout, STDOUT_FILENO);
			dup2(Old_Stderr, STDERR_FILENO);
			message(1, "Execution of '%s' failed in thread %d: %s\n", Command, UniqueThreadID, strerror(errno));
		}
		exit(EXIT_FAILURE);
	} else {
		/* Close writing end of pipe */
		close(Modus[1]);

		/* Parent process */
		((*Processes)[CurrentProcess]).PID = PID;
		((*Processes)[CurrentProcess]).Valid = 1;
		message(2, "Forked off child '%s %s' with PID %d (unique thread ID %d).\n", Command, Parameter, PID, UniqueThreadID);

	}
	Data->Active = 1;		/* Start redirection */
	sigprocmask(SIG_UNBLOCK, &Mask, NULL);
	
	return CurrentProcess;
}

int WaitFor(int ProcID, struct Process **Processes) {
	double Before, After;
	int Return;
	
	Before = GetTime();
	while (((*Processes)[ProcID]).Running == 1) {
/*		printf("Process %d state: pid=%d running=%d return=%d\n", ProcID, Processes[ProcID].PID, Processes[ProcID].Running, Processes[ProcID].Status); */
		usleep(100 * 1000);	/* 100ms */
	}
	After = GetTime();
	
	Return = -1;
	if (WIFSIGNALED((*Processes)[ProcID].Status)) {
		char *Msg = SignalMessage((*Processes)[ProcID].Status);
		Return = -WTERMSIG((*Processes)[ProcID].Status);
		if (Msg != NULL) {	
			message(2, "Waited for thread %d %.2f seconds, signalled by signal %s.\n", ProcID, After - Before, Msg);
		} else {
			message(2, "Waited for thread %d %.2f seconds, signalled by signal number %d.\n", ProcID, After - Before, -Return);
		}
	} else {
		if (WIFEXITED((*Processes)[ProcID].Status)) {
			Return = WEXITSTATUS((*Processes)[ProcID].Status);
			message(2, "Waited for thread %d %.2f seconds, exited with status %d.\n", ProcID, After - Before, Return);
		} else {
			message(2, "Waited for thread %d %.2f seconds, exited stange (status %x).\n", ProcID, After - Before, (*Processes)[ProcID].Status);
		}
	}

	/* Also join the according thread ID */
	pthread_join((*Processes)[ProcID].ThreadID, NULL);

	return Return;
}

