/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2006 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/wait.h>
#include <time.h>
#include <pthread.h>
#include <sys/time.h>

#include "String.h"
#include "Mutex.h"
#include "Parser.h"
#include "CommandLine.h"
#include "Threads.h"
#include "Processes.h"
#include "sambascanner.h"

static struct Process *Processes = NULL;
static int NumberProcesses = 0;

void SIGCHLD_Handler(int Signal) {
	int i;
	int Errno_Backup;
	pid_t PID;
	Errno_Backup = errno;
	for (i = 0; i < NumberProcesses; i++) {
		if (Processes[i].Valid==1) {
			PID = waitpid(Processes[i].PID, &Processes[i].Status, WNOHANG);
/*			fprintf(stderr, "Handling SIGCHLD [%d] => PID = %d, result = %d...\n", i, Processes[i].PID, PID); */
			if (PID > 0) {
				/* Child really exited */
				Processes[i].Running = 0;
			}
		} else {
/*			fprintf(stderr, "Invalid process %d ignored!\n", i); */
		}
	}
	errno = Errno_Backup;
}

void Install_SigHandler() {
	struct sigaction Action;
	
	Action.sa_handler = SIGCHLD_Handler;
	Action.sa_flags = SA_RESTART;
	sigemptyset(&Action.sa_mask);
	
	sigaction(SIGCHLD, &Action, NULL);
}

int main(int argc, char **argv) {
	StringList Sharelist;
	StringList Hostlist;
	StringList Resultfilelist;
	static struct Process_Watcher_Thread_InputData Process_Watcher_Data;
	
	Init_StringList(&Sharelist);
	Init_StringList(&Hostlist);
	Init_StringList(&Resultfilelist);

	if (!Process_Parameters(argc, argv, &Hostlist)) {
		return 1;
	};
	
	Init_Semaphores(Options.ThreadCount);
	Install_SigHandler();
	{
		pthread_t ProcessWatcher;
		pthread_attr_t Attrs;
		pthread_attr_init(&Attrs);
		pthread_attr_setdetachstate(&Attrs, PTHREAD_CREATE_DETACHED);
	
		Process_Watcher_Data.Running = 1;
		Process_Watcher_Data.NumberProcesses = &NumberProcesses;
		Process_Watcher_Data.Processes = &Processes;

		pthread_create(&ProcessWatcher, &Attrs, Process_Watcher_Thread, &Process_Watcher_Data);
	}


	if (Options.Workgroup[0] == 0) {
		int ProcID;
		int Result;
		/* Workgroup was not set in options */
		message(3, "STAGE: Now entering stage one.\n");
		message(1, "Looking up workgroup...\n");
		ProcID = Exec_Timeout(Options.NmbLookup_Executable, "-M -S -- -", 10.0, 3.0, FILE_WORKGROUPINFO, -1, &Processes, &NumberProcesses);
		Result = WaitFor(ProcID, &Processes);
		if (Result!=0) {
			message(1, "Process terminated at looking up workgroup.\n");
			return 1;
		}
		if (Parse_Groupinfo(FILE_WORKGROUPINFO, Options.Workgroup)!=0) {
			message(1, "Process terminated at parsing workgroup info list.\n");
			return 2;
		};
		message(2, "Returned workgroup: '%s'\n", Options.Workgroup);
		if (!Options.PreserveTemp) unlink(FILE_WORKGROUPINFO);
		
		if (strlen(Options.Workgroup)==0) {
			fprintf(stderr, "No workgroup found, aborting!\n");
			fprintf(stderr, "Try setting the workgroup manually.\n");
			return 3;
		}
	} else {
		message(3, "STAGE: Stage one skipped.\n");
	}
	
	{
		int ProcID;
		int Result;
		char Buffer[BUFSIZE_SHORT];
		message(3, "STAGE: Now entering stage two.\n");
		snprintf(Buffer, sizeof(Buffer), "-M -S %s", Options.Workgroup);
		message(1, "Retrieving master browsers in workgroup '%s'...\n", Options.Workgroup);
		ProcID = Exec_Timeout(Options.NmbLookup_Executable, Buffer, 10.0, 3.0, FILE_MASTERBROWSER, -1, &Processes, &NumberProcesses);
		Result = WaitFor(ProcID, &Processes);
		if (Result!=0) {
			message(1, "Process terminated at retrieving master browsers.\n");
			return 1;
		}
		if (Parse_Masterbrowser(FILE_MASTERBROWSER, &Hostlist) != 0) {
			message(1, "Process terminated at parsing master browser list.\n");
			return 1;
		}
		if (!Options.PreserveTemp) unlink(FILE_MASTERBROWSER);
	}
	
	message(3, "STAGE: Now entering stage three.\n");
	{
		int i;
		int NumberKilled, NumberOkay;
		int Hostlookup_Processes_Succeeded = 0;
		int Hostlookup_Processes_Failed = 0;
		struct Process_Host_Thread_InputData *Process_Host_Data;
		pthread_mutex_t NumberMutex;

		pthread_mutex_init(&NumberMutex, NULL);
		Process_Host_Data = (struct Process_Host_Thread_InputData*)malloc(sizeof(struct Process_Host_Thread_InputData) * Hostlist.Entries);

		NumberKilled = 0;
		NumberOkay = 0;
		ThreadCnt_Set(Options.ThreadCount);
		for (i = 0; i < Hostlist.Entries; i++) {
			pthread_t TempThread;

			Process_Host_Data[i].Hostname = Hostlist.Entry[i];
			Process_Host_Data[i].Hostlist = &Hostlist;
			Process_Host_Data[i].Sharelist = &Sharelist;
			Process_Host_Data[i].NumberMutex = &NumberMutex;
			Process_Host_Data[i].Hostlookup_Processes_Succeeded = &Hostlookup_Processes_Succeeded;
			Process_Host_Data[i].Hostlookup_Processes_Failed = &Hostlookup_Processes_Failed;
			Process_Host_Data[i].Processes = &Processes;
			Process_Host_Data[i].NumberProcesses = &NumberProcesses;

			ThreadCnt_Inc();
			pthread_create(&TempThread, NULL, Process_Host_Thread, Process_Host_Data + i);
		}
		message(3, "Waiting for threads to return home...\n");
		for (i = 0; i < Options.ThreadCount; i++) ThreadCnt_Inc();	/* Wait until all threads are finished */
		message(2, "%d processes killed (%.1f%%), %d returned sucessfully (%.1f%%).\n", 
				Hostlookup_Processes_Failed, 
				100.0 * Hostlookup_Processes_Failed / (Hostlookup_Processes_Failed + Hostlookup_Processes_Succeeded), 
				Hostlookup_Processes_Succeeded, 
				100.0 * Hostlookup_Processes_Succeeded / (Hostlookup_Processes_Failed + Hostlookup_Processes_Succeeded)
			);
		message(2, "Total hosts found: %d\n", Hostlist.Entries);

		/* Now that all processes are terminated, allocated data can be freed again */
		free(Process_Host_Data);
	}

	/*
	{
		int i;
		fprintf(stderr, "Number of hosts in list: %d\n", NumberHosts);
		for (i=0; i<NumberHosts; i++) {
			fprintf(stderr, "%3d: %s\n", i+1, Hostlist[i]);
		}
		fprintf(stderr, "\n");
	}
	*/

	/* Now try to reveal all share content on the found shares */
	message(3, "STAGE: Now entering stage four.\n");
	{
		int i;
		int UniqueThreadID=0;
		struct Process_Share_Thread_InputData *Process_Share_Data;
		
		ThreadCnt_Set(Options.ThreadCount);
		message(1, "Revealing share content of %d %s in %d %s...\n", Sharelist.Entries, SingularPlural("share", "shares", Sharelist.Entries), Hostlist.Entries, SingularPlural("host", "hosts", Sharelist.Entries));
		
		Process_Share_Data = (struct Process_Share_Thread_InputData*)malloc(sizeof(struct Process_Share_Thread_InputData) * Sharelist.Entries);
		for (i = 0; i < Sharelist.Entries; i++) {
			pthread_t TempThread;
			ThreadCnt_Inc();
			Process_Share_Data[i].Sharename = Sharelist.Entry[i];
			Process_Share_Data[i].UniqueThreadID = UniqueThreadID;
			Process_Share_Data[i].Resultfilelist = &Resultfilelist;
			Process_Share_Data[i].Processes = &Processes;
			Process_Share_Data[i].NumberProcesses = &NumberProcesses;
			pthread_create(&TempThread, NULL, Process_Share_Thread, Process_Share_Data + i);
			UniqueThreadID++;
		}
	
		/* Now wait for the last few threads to come home */

		message(3, "Waiting for threads to return home...\n");
		for (i = 0; i < Options.ThreadCount; i++) ThreadCnt_Inc();	/* Wait until all threads are finished */
		message(3, "...all done.\n");
	
		/* Then, free memory */
		free(Process_Share_Data);
		Destroy_StringList(&Sharelist);
	}

	{
		int i;
		FILE *f, *g;
		char Line[BUFSIZE];
	
		f = fopen(Options.Outfile, "w");
		if (f == NULL) {
			message(1, "%s: Error writing outfile '%s' for writing (%s)\n", argv[0], Options.Outfile, strerror(errno));
			return 1;
		}
		for (i = 0; i < Resultfilelist.Entries; i++) {
			message(2, "Merging file '%s'...\n", Resultfilelist.Entry[i]);
			g = fopen(Resultfilelist.Entry[i], "r");
			if (g != NULL) {
				while (fgets(Line, BUFSIZE, g) != NULL) {
					if (strlen(Line) > 0) {
						if (Line[strlen(Line) - 1] == '\n') Line[strlen(Line) - 1] = '\0';
						if (strncmp(Line, "smb://", 6) == 0) fprintf(f, "%s\n", Line);
					}
				}
				fclose(g);
			}
			if (!Options.PreserveTemp) unlink(Resultfilelist.Entry[i]);
		}
		fclose(f);
	}

	/* Wait for process watcher to die */
	message(3, "STAGE: Now entering stage five.\n");
	
	/* Tell process watcher to exit */
	Process_Watcher_Data.Running = 0;

	/* And wait for its acknowledgement */
	while (Process_Watcher_Data.Running != -1) {
		usleep(100 * 1000);
	}

	Destroy_StringList(&Hostlist);
	Destroy_StringList(&Resultfilelist);

	return 0;
}
